#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <pthread.h>
#include <time.h>

#include <SDL.h>

static void die(const char* message, ...) {
  va_list ap;
  va_start(ap, message);
  vfprintf(stderr, message, ap);
  exit(1);
};

#define defdie(name, error) \
  void name ## _die (const char* function) { \
    die("%s failed: %s\n", function, error); \
  };

defdie(sdl, SDL_GetError());
defdie(posix, strerror(errno));

#undef defdie

static void* xmalloc(size_t size) {
  void* result = malloc(size);
  if (!result) posix_die("malloc");
  return result;
};

static void time_sub(struct timespec* a, struct timespec* b, struct timespec* result) {
  result->tv_sec = a->tv_sec - b->tv_sec;
  result->tv_nsec = a->tv_nsec - b->tv_nsec;
  if (result->tv_nsec < 0) {
    --result->tv_sec;
    result->tv_nsec += 1000000000;
  } else if (result->tv_nsec > 1000000000) {
    ++result->tv_sec;
    result->tv_nsec -= 1000000000;
  };
};

static long gettime() {
  struct timespec tp;
  if (clock_gettime(CLOCK_MONOTONIC, &tp) != 0) posix_die("clock_gettime");
  return tp.tv_nsec;
};

static void sdl_set_color(
    SDL_Renderer* renderer, Uint8 red, Uint8 green, Uint8 blue, Uint8 alpha
) {
  if (SDL_SetRenderDrawColor(renderer, red, green, blue, alpha))
    sdl_die("SDL_SetRenderDrawColor");
};

typedef struct point {
  int x;
  int y;
} Point;

typedef struct segment {
  Point           cell;
  struct segment* prev;
  struct segment* next;
} Segment;

static Segment* create_segment(Point* cell, Segment* prev, Segment* next) {
  Segment* result = xmalloc(sizeof(Segment));
  result->cell = *cell;

  result->prev = prev;
  if (prev) prev->next = result;
  result->next = next;
  if (next) next->prev = result;

  return result;
};

static void insert_segment_after(Segment* anchor, Segment* segment) {
  segment->prev = anchor;
  segment->next = anchor->next;
  anchor->next = segment;
};

static void insert_segment_before(Segment* anchor, Segment* segment) {
  segment->next = anchor;
  segment->prev = anchor->prev;
  anchor->prev = segment;
};

static void unlink_segment(Segment* segment) {
  if (segment->prev) {
    segment->prev->next = segment->next;
    segment->prev = NULL;
  };

  if (segment->next) {
    segment->next->prev = segment->prev;
    segment->next = NULL;
  };
};

typedef enum direction {
  dirUp    = 0,
  dirLeft  = 1,
  dirRight = 2,
  dirDown  = 3
} Direction;

typedef struct snake {
  Segment*        head;
  Segment*        tail;
  Direction       direction[8];
  int             dir_get;
  int             dir_set;
  pthread_mutex_t mutex;
  unsigned        length;
  int             growing;
} Snake;

#define FIELD_WIDTH  15
#define FIELD_HEIGHT 15
#define CELL_SIZE    30

typedef char Field[FIELD_WIDTH][FIELD_HEIGHT];

static void grow_snake(Snake* snake, Field field, Point* cell) {
  snake->tail = create_segment(cell, snake->tail, NULL);
  field[cell->x][cell->y] = 1;
  ++snake->length;
};

static Snake* create_snake(Field field) {
  Point cell;
  cell.x = FIELD_WIDTH  / 2;
  cell.y = FIELD_HEIGHT / 2;

  Segment* head = create_segment(&cell, NULL, NULL);
  field[cell.x][cell.y] = 1;

  Snake* snake = xmalloc(sizeof(Snake));
  memset(snake, 0, sizeof(*snake));
  snake->head = head;
  snake->tail = head;
  if (pthread_mutex_init(&snake->mutex, NULL) != 0)
    posix_die("pthread_mutex_init");

  for (int i = 1; i < 3; ++i) {
    ++cell.x;
    grow_snake(snake, field, &cell);
  };

  return snake;
};

static void next_cell(Point* cell, Direction direction) {
  switch (direction) {
    case dirUp:
      --cell->y;
      break;
    case dirDown:
      ++cell->y;
      break;
    case dirLeft:
      --cell->x;
      break;
    case dirRight:
      ++cell->x;
      break;
  };
};

static inline int next_snake_direction(Snake* snake, int dir) {
  if (++dir >= sizeof(snake->direction) / sizeof(*snake->direction)) dir = 0;
  return dir;
};

static void set_snake_direction(Snake* snake, Direction direction) {
  if (direction == snake->direction[snake->dir_set]
      || direction == 3 - snake->direction[snake->dir_set])
    return;
  int i = next_snake_direction(snake, snake->dir_set);
  if (i == snake->dir_get) return;
  snake->dir_set = i;
  snake->direction[i] = direction;
};

static Direction get_snake_direction(Snake* snake) {
  if (snake->dir_get != snake->dir_set)
    snake->dir_get = next_snake_direction(snake, snake->dir_get);
  return snake->direction[snake->dir_get];
};

static void move_snake(Snake* snake, Field field) {
  Direction dir = get_snake_direction(snake);

  if (pthread_mutex_lock(&snake->mutex) != 0) posix_die("pthread_mutex_lock");

  Segment* s;
  if (snake->growing) {
    s = create_segment(&snake->head->cell, NULL, snake->head);
    ++snake->length;
    snake->growing = 0;
  } else {
    s = snake->tail;
    snake->tail = snake->tail->prev;
    unlink_segment(s);
    insert_segment_before(snake->head, s);
    --field[s->cell.x][s->cell.y];
    s->cell = snake->head->cell;
  };

  next_cell(&s->cell, dir);
  if (s->cell.x >= 0
      && s->cell.x < FIELD_WIDTH
      && s->cell.y >= 0
      && s->cell.y < FIELD_HEIGHT)
    ++field[s->cell.x][s->cell.y];
  snake->head = s;

  if (pthread_mutex_unlock(&snake->mutex) != 0) posix_die("pthread_mutex_unlock");
};

static void render_cell(SDL_Renderer* renderer, Point* cell) {
  SDL_Rect rect;
  rect.x = cell->x * CELL_SIZE;
  rect.y = cell->y * CELL_SIZE;
  rect.w = CELL_SIZE;
  rect.h = CELL_SIZE;
  if (SDL_RenderFillRect(renderer, &rect) != 0) sdl_die("SDL_RenderFillRect");
};

static void render_snake(Snake* snake, SDL_Renderer* renderer) {
  if (pthread_mutex_lock(&snake->mutex) != 0) posix_die("pthread_mutex_lock");

#if 0
  Segment* segment = snake->head;
  sdl_set_color(renderer, 0x00, 0x00, 0xff, 0xff);
  render_cell(renderer, &segment->cell);

  sdl_set_color(renderer, 0xff, 0xff, 0xff, 0xff);
  for (segment = segment->next; segment != snake->tail; segment = segment->next)
    render_cell(renderer, &segment->cell);
  
  sdl_set_color(renderer, 0x00, 0xff, 0x00, 0xff);
  render_cell(renderer, &snake->tail->cell);
#endif

  const char colors[2][3] = {
    { 0xd2, 0x69, 0x1e },
    { 0x00, 0x80, 0x00 }
  };
  double step[3];
  for (int i = 0; i < 3; ++i)
    step[i] = (double)(colors[1][i] - colors[0][i]) / snake->length;

  char color[3];
  for (int i = 0; i < 3; ++i) color[i] = colors[0][i];
  for (Segment* segment = snake->head; segment; segment = segment->next) {
    sdl_set_color(renderer, color[0], color[1], color[2], 0xff);
    render_cell(renderer, &segment->cell);
    for (int i = 0; i < 3; ++i) color[i] += step[i];
  };

  if (pthread_mutex_unlock(&snake->mutex) != 0) posix_die("pthread_mutex_unlock");
};

static void tick_loop(int tick_length /* ms */, int (*body)(void*), void* arg) {
  struct timespec length;
  length.tv_sec = tick_length / 1000;
  length.tv_nsec = (tick_length % 1000) * 1000000;
  while (1) {
    struct timespec start;
    if (clock_gettime(CLOCK_MONOTONIC, &start) != 0) posix_die("clock_gettime");

    if (!body(arg)) return;

    struct timespec end;
    if (clock_gettime(CLOCK_MONOTONIC, &end) != 0) posix_die("clock_gettime");

    struct timespec diff;
    time_sub(&end, &start, &diff);

    struct timespec delay;
    time_sub(&length, &diff, &delay);
    if (delay.tv_sec >= 0 && delay.tv_nsec >= 0)
      while (nanosleep(&delay, &delay) == -1 && errno == EINTR);
  };
};

typedef struct world {
  SDL_Renderer* renderer;
  Snake*        snake;
  Point         snack;
  Field         field;
} World;

static void reset_snack(World* world) {
  int n = rand() % (FIELD_WIDTH * FIELD_HEIGHT - world->snake->length);
  char* f = (char*)world->field;
  while (n > 0) if (!*f++) --n;
  world->snack.x = (f - (char*)world->field) % FIELD_WIDTH;
  world->snack.y = (f - (char*)world->field) / FIELD_WIDTH;
};

static void render_snack(Point* snack, SDL_Renderer* renderer) {
  sdl_set_color(renderer, 255, 0, 0, 255);
  render_cell(renderer, snack);
};

static int tick_world(World* world) {
  move_snake(world->snake, world->field);

  Point* head = &world->snake->head->cell;

  if (head->x == world->snack.x && head->y == world->snack.y) {
    world->snake->growing = 1;
    reset_snack(world);
  };

  if (head->x < 0 || head->x >= FIELD_WIDTH || head->y < 0 || head->y >= FIELD_HEIGHT)
    return 0;

  if (world->field[head->x][head->y] > 1)
    return 0;

  return 1;
};

static void* world_main(World* world) {
  tick_loop(100, (int (*)(void*))tick_world, world);
  return NULL;
};

static int render_world(World* world) {
  SDL_Event event;
  int quit = 0;

  while (SDL_PollEvent(&event))
    switch (event.type) {
      case SDL_QUIT:
        return 0;

      case SDL_KEYDOWN:
        switch (event.key.keysym.sym) {
          case SDLK_UP:
            set_snake_direction(world->snake, dirUp);
            break;
          case SDLK_DOWN:
            set_snake_direction(world->snake, dirDown);
            break;
          case SDLK_LEFT:
            set_snake_direction(world->snake, dirLeft);
            break;
          case SDLK_RIGHT:
            set_snake_direction(world->snake, dirRight);
            break;
        };
        break;
    };

  sdl_set_color(world->renderer, 0, 0, 0, 255);
  if (SDL_RenderClear(world->renderer) != 0)
    sdl_die("SDL_RenderClear");

  render_snake(world->snake, world->renderer);
  render_snack(&world->snack, world->renderer);

  SDL_RenderPresent(world->renderer);

  return 1;
};

static void init_rng() {
  FILE* urandom = fopen("/dev/urandom", "r");
  if (!urandom) die("Cannot open /dev/urandom: %s", strerror(errno));
  unsigned seed;
  if (fread(&seed, sizeof(seed), 1, urandom) != 1)
    die(
        "Failed to read %zd bytes from /dev/urandom: %s\n",
        sizeof(seed),
        strerror(errno)
    );
  if (fclose(urandom) != 0)
    fprintf(stderr, "Failed to close /dev/urandom: %s\n", strerror(errno));
  srand(seed);
};

int main() {
  init_rng();

  SDL_Init(SDL_INIT_VIDEO);

  SDL_Window* window = SDL_CreateWindow(
      "Snake",
      SDL_WINDOWPOS_UNDEFINED,
      SDL_WINDOWPOS_UNDEFINED,
      FIELD_WIDTH  * CELL_SIZE,
      FIELD_HEIGHT * CELL_SIZE,
      0
  );
  if (!window) sdl_die("SDL_CreateWindow");

  World world;
  memset(world.field, 0, sizeof(world.field));

  world.renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_ACCELERATED);
  if (!world.renderer) sdl_die("SDL_CreateRenderer");

  world.snake = create_snake(world.field);
  reset_snack(&world);

  pthread_t world_thread;
  if (pthread_create(&world_thread, NULL, (void* (*)(void*))world_main, &world) != 0)
    posix_die("pthread_create");

  tick_loop(1000 / 60, (int (*)(void*))render_world, &world);

  return 0;
};
